from tensorflow.keras.models import Sequential, save_model
from tensorflow.keras.layers import Input, Conv1D, Conv1DTranspose, Dense
from tensorflow.keras.constraints import max_norm
from tensorflow.keras.optimizers import SGD, RMSprop, Adam, Adamax, Adagrad, Adadelta
from keras.models import Model
import matplotlib.pyplot as plt
import numpy as np
import math
from scipy import signal

from server.sl8 import FileBase
from extra import design_filters

contaminated_file_list = ['./iec_files/1615632656_EyeOpen_1_Other.iec',
                          './iec_files/1618321042_EyeOpen_6_sina_izani.iec',
                          './iec_files/1618747412_EyeOpen_7_melika_hoseinzadeh.iec',
                          './iec_files/1619683416_EyeOpen_9_khosravi.iec',
                          './iec_files/1620309313_EyeOpen_12_mr_mohsen_shahbaz_beigi.iec',
                          './iec_files/1620576311_EyeOpen_14_mahan_endallah.iec',
                          './iec_files/1621949668_EyeOpen_16_diyana_ebrahimi_0.iec',
                          './iec_files/1623149547_EyeOpen_1_Other_3.iec',
                          './iec_files/1623154705_EyeOpen_1_Other_3.iec',
                          './iec_files/1623597049_EyeOpen_18_anita_mirzayi_0.iec',
                          './iec_files/1624281526_EyeOpen_1_Other_0.iec',
                          './iec_files/1624281967_EyeOpen_1_Other_0.iec'
                          ]
all_contaminated, all_clean = [], []
for contaminated_file in contaminated_file_list:
    clean_file = 'blink_removed_iec_files/' + contaminated_file.split('/')[-1]

    labeled_filename = contaminated_file.replace('iec_files', 'label_results').replace('iec', 'txt')

    # make data
    iec = FileBase()
    iec.load(contaminated_file)
    contaminated_data = iec.get_brain_raw_data()
    sr = iec.get_brain_sample_rate()
    gain = iec.get_gain()

    filters = design_filters(sr)
    contaminated_data = signal.filtfilt(filters['notch1'][0], filters['notch1'][1], contaminated_data)
    contaminated_data = signal.filtfilt(filters['notch2'][0], filters['notch2'][1], contaminated_data)
    contaminated_data = signal.filtfilt(filters['bandpass'][0], filters['bandpass'][1], contaminated_data)

    iec2 = FileBase()
    iec2.load(clean_file)
    clean_data = iec2.get_brain_raw_data()
    clean_sr = iec2.get_brain_sample_rate()
    clean_gain = iec2.get_gain()

    # from blink_remove import eegplot
    # eegplot(contaminated_data, sr, 'bad')
    # eegplot(clean_data, clean_sr, 'good')

    # with open(labeled_filename) as f:
    #     content = f.readlines()
    # # you may also want to remove whitespace characters like `\n` at the end of each line
    # content = [x.strip()[1:-1] for x in content]
    #
    # arr = np.zeros((len(content), 4), dtype=int)
    # for i in range(len(content)):
    #     temp = content[i].split(',')
    #     arr[i, 0] = int(temp[0])
    #     arr[i, 1] = int(temp[1])
    #     arr[i, 2] = int(temp[2])
    #     arr[i, 3] = int(temp[3])
    # del content
    arr = np.loadtxt(labeled_filename, dtype=int)

    contaminated_index = arr[arr[:, -1] == 1]
    contaminated_parts = []
    clean_parts = []
    for i in range(contaminated_index.shape[0]):
        contaminated_parts.append(
            contaminated_data[contaminated_index[i, 0], contaminated_index[i, 1]: contaminated_index[i, 2]])
        clean_parts.append(clean_data[contaminated_index[i, 0], contaminated_index[i, 1]: contaminated_index[i, 2]])
        # my_data = my_data / gain
        # my_data -= np.mean(my_data)
        # my_data = signal.detrend(my_data)

    min_len_contaminated_parts = len(min(contaminated_parts, key=len))
    min_len_clean_parts = len(min(clean_parts, key=len))

    for i in range(len(contaminated_parts)):
        if len(contaminated_parts[i]) > min_len_contaminated_parts:
            contaminated_parts[i] = contaminated_parts[i][:-1]
    for i in range(len(clean_parts)):
        if len(clean_parts[i]) > min_len_clean_parts:
            clean_parts[i] = clean_parts[i][:-1]
    contaminated_parts = np.array(contaminated_parts)
    clean_parts = np.array(clean_parts)

    all_contaminated.append(contaminated_parts)
    all_clean.append(clean_parts)

contaminated_parts = all_contaminated[0]
for i in range(0, len(all_contaminated) - 1):
    contaminated_parts = np.concatenate((contaminated_parts, all_contaminated[i + 1]), axis=0)
clean_parts = all_clean[0]
for i in range(0, len(all_clean) - 1):
    clean_parts = np.concatenate((clean_parts, all_clean[i + 1]), axis=0)

# for i in range(0, len(clean_parts)):
#     # Get the sample and the reconstruction
#     original = contaminated_parts[i]
#     pure = clean_parts[i]
#     # Matplotlib preparations
#     fig, axes = plt.subplots(1, 2)
#     # Plot sample and reconstruciton
#     axes[0].plot(original)
#     axes[0].set_title('Noisy waveform')
#     axes[1].plot(pure)
#     axes[1].set_title('Pure waveform')
#     plt.show()

# train autoencoder
desired_model = 'conv'

if desired_model == 'conv':
    # Model configuration
    input_shape = (100, 1)
    batch_size = 128
    no_epochs = 100
    train_test_split = 0.2
    validation_split = 0.2
    verbosity = 1
    max_norm_value = 2.0

    # Reshape data
    y_val_noisy_r = []
    y_val_pure_r = []
    for i in range(0, len(contaminated_parts)):
        noisy_sample = contaminated_parts[i]
        pure_sample = clean_parts[i]
        noisy_sample = (noisy_sample - np.min(noisy_sample)) / (np.max(noisy_sample) - np.min(noisy_sample))
        pure_sample = (pure_sample - np.min(pure_sample)) / (np.max(pure_sample) - np.min(pure_sample))
        # noisy_sample = noisy_sample / np.max(noisy_sample)
        # pure_sample = pure_sample / np.max(pure_sample)
        y_val_noisy_r.append(noisy_sample)
        y_val_pure_r.append(pure_sample)
    y_val_noisy_r = np.array(y_val_noisy_r)
    y_val_pure_r = np.array(y_val_pure_r)
    noisy_input = y_val_noisy_r.reshape((y_val_noisy_r.shape[0], y_val_noisy_r.shape[1], 1))
    pure_input = y_val_pure_r.reshape((y_val_pure_r.shape[0], y_val_pure_r.shape[1], 1))

    # Train/test split
    percentage_training = math.floor((1 - train_test_split) * len(noisy_input))
    noisy_input, noisy_input_test = noisy_input[:percentage_training], noisy_input[percentage_training:]
    pure_input, pure_input_test = pure_input[:percentage_training], pure_input[percentage_training:]

    # Create the model
    model = Sequential()
    model.add(Conv1D(128, kernel_size=3, kernel_constraint=max_norm(max_norm_value), activation='relu',
                     kernel_initializer='he_uniform', input_shape=input_shape))
    # model.add(Conv1D(64, kernel_size=3, kernel_constraint=max_norm(max_norm_value), activation='relu',
    #                  kernel_initializer='he_uniform'))
    model.add(Conv1D(32, kernel_size=3, kernel_constraint=max_norm(max_norm_value), activation='relu',
                     kernel_initializer='he_uniform'))
    model.add(Conv1DTranspose(32, kernel_size=3, kernel_constraint=max_norm(max_norm_value), activation='relu',
                              kernel_initializer='he_uniform'))
    # model.add(Conv1DTranspose(64, kernel_size=3, kernel_constraint=max_norm(max_norm_value), activation='relu',
    #                           kernel_initializer='he_uniform'))
    model.add(Conv1DTranspose(128, kernel_size=3, kernel_constraint=max_norm(max_norm_value), activation='relu',
                              kernel_initializer='he_uniform'))
    model.add(
        Conv1D(1, kernel_size=3, kernel_constraint=max_norm(max_norm_value), activation='sigmoid', padding='same'))

    model.summary()

    # Compile and fit data
    # model.compile(optimizer='adam', loss='binary_crossentropy')
    model.compile(optimizer=Adam(learning_rate=0.001), loss='binary_crossentropy')

    # model.fit(noisy_input, pure_input,
    #           epochs=no_epochs,
    #           batch_size=batch_size,
    #           validation_split=validation_split,
    #           verbose=verbosity)

    history = model.fit(x=noisy_input, y=pure_input,
                        epochs=no_epochs,
                        batch_size=batch_size,
                        verbose=verbosity,
                        shuffle=True,
                        validation_data=(noisy_input_test, pure_input_test))

    # "Loss"
    plt.plot(history.history['loss'])
    plt.plot(history.history['val_loss'])
    plt.title('model loss')
    plt.ylabel('loss')
    plt.xlabel('epoch')
    plt.legend(['train', 'validation'], loc='upper left')
    plt.show()

    # Generate reconstructions
    num_reconstructions = 20
    samples = noisy_input_test[:num_reconstructions]
    # samples = pure_input_test[:num_reconstructions]
    reconstructions = model.predict(samples)

    # Plot reconstructions
    for i in np.arange(0, num_reconstructions):
        # Prediction index
        prediction_index = i + percentage_training
        # Get the sample and the reconstruction
        original = contaminated_parts[prediction_index]
        pure = clean_parts[prediction_index]
        reconstruction = np.array(reconstructions[i])
        # # Matplotlib preparations
        # fig, axes = plt.subplots(1, 3)
        # # Plot sample and reconstruciton
        # axes[0].plot(original)
        # axes[0].set_title('Noisy waveform')
        # axes[1].plot(pure)
        # axes[1].set_title('Pure waveform')
        # axes[2].plot(reconstruction)
        # axes[2].set_title('Conv Autoencoder Denoised waveform')
        # plt.show()

        # Matplotlib preparations
        fig, axes = plt.subplots(2, 3)
        # Plot sample and reconstruciton
        axes[0, 0].plot(original/max(abs(original)))
        axes[0, 0].set_title('Noisy waveform')
        axes[0, 1].plot(pure/max(abs(pure)))
        axes[0, 1].set_title('Pure waveform')
        axes[0, 2].plot(reconstruction/max(abs(reconstruction)))
        axes[0, 2].set_title('Denoised waveform')

        f, pxx = signal.welch(original, fs=256,
                              window=signal.windows.tukey(100, sym=False, alpha=.17),
                              nperseg=100, noverlap=int(100 * 0.75),
                              nfft=2 * 256, return_onesided=True,
                              scaling='spectrum', axis=-1, average='mean')
        axes[1, 0].plot(f, pxx)
        axes[1, 0].set_title('Noisy PSD')
        axes[1, 0].set_xlim((0, 30))
        f, pxx = signal.welch(pure, fs=256,
                              window=signal.windows.tukey(100, sym=False, alpha=.17),
                              nperseg=100, noverlap=int(100 * 0.75),
                              nfft=2 * 256, return_onesided=True,
                              scaling='spectrum', axis=-1, average='mean')
        axes[1, 1].plot(f, pxx)
        axes[1, 1].set_title('Pure PSD')
        axes[1, 1].set_xlim((0, 30))
        f, pxx = signal.welch(reconstruction.T.squeeze(), fs=256,
                              window=signal.windows.tukey(100, sym=False, alpha=.17),
                              nperseg=100, noverlap=int(100 * 0.75),
                              nfft=2 * 256, return_onesided=True,
                              scaling='spectrum', axis=-1, average='mean')
        axes[1, 2].plot(f, pxx)
        axes[1, 2].set_title('Denoised PSD')
        axes[1, 2].set_xlim((0, 30))

        plt.show()

    # model.save('conv_model.h5')

elif desired_model == 'dense':

    # contaminated_parts = np.vstack([contaminated_parts] * 50)
    # clean_parts = np.vstack([clean_parts] * 50)

    # Reshape data
    y_val_noisy_r = []
    y_val_pure_r = []
    for i in range(0, len(contaminated_parts)):
        noisy_sample = contaminated_parts[i]
        pure_sample = clean_parts[i]
        noisy_sample = (noisy_sample - np.min(noisy_sample)) / (np.max(noisy_sample) - np.min(noisy_sample))
        pure_sample = (pure_sample - np.min(pure_sample)) / (np.max(pure_sample) - np.min(pure_sample))
        # noisy_sample = noisy_sample / np.max(noisy_sample)
        # pure_sample = pure_sample / np.max(pure_sample)
        y_val_noisy_r.append(noisy_sample)
        y_val_pure_r.append(pure_sample)
    noisy_input = np.array(y_val_noisy_r)
    pure_input = np.array(y_val_pure_r)

    # Train/test split
    percentage_training = math.floor((1 - 0.2) * len(noisy_input))
    noisy_input, noisy_input_test = noisy_input[:percentage_training], noisy_input[percentage_training:]
    pure_input, pure_input_test = pure_input[:percentage_training], pure_input[percentage_training:]

    # Input image
    input_img = Input(shape=(100,))
    # encoded and decoded layer for the autoencoder
    encoded = Dense(units=128, activation='relu')(input_img)
    encoded = Dense(units=64, activation='relu')(encoded)
    encoded = Dense(units=32, activation='relu')(encoded)
    decoded = Dense(units=64, activation='relu')(encoded)
    decoded = Dense(units=128, activation='relu')(decoded)
    decoded = Dense(units=100, activation='sigmoid')(decoded)

    # Building autoencoder
    autoencoder = Model(input_img, decoded)
    # extracting encoder
    encoder = Model(input_img, encoded)

    # compiling the autoencoder
    autoencoder.compile(optimizer='adam', loss='binary_crossentropy')  # , metrics=['accuracy'])
    # Fitting the noise trained data to the autoencoder

    # history = autoencoder.fit(noisy_input, pure_input,
    #                           epochs=1000,
    #                           batch_size=128,
    #                           validation_split=0.2,
    #                           verbose=2)

    history = autoencoder.fit(x=noisy_input, y=pure_input,
                              epochs=100,
                              batch_size=64,
                              verbose=2,
                              shuffle=True,
                              validation_data=(noisy_input_test, pure_input_test))

    # "Loss"
    plt.plot(history.history['loss'])
    plt.plot(history.history['val_loss'])
    plt.title('model loss')
    plt.ylabel('loss')
    plt.xlabel('epoch')
    plt.legend(['train', 'validation'], loc='upper left')
    plt.show()

    # reconstructing the image from autoencoder and encoder
    encoded_imgs = encoder.predict(noisy_input_test)
    predicted = autoencoder.predict(noisy_input_test)

    # # plotting the noised image, encoded image and the reconstructed image
    # plt.figure(figsize=(20, 2))
    # for i in range(10):
    #     # display original images
    #
    #     ax = plt.subplot(4, 20, i + 1)
    #     plt.plot(pure_input[i])
    #
    #     # display noised images
    #     ax = plt.subplot(4, 20, i + 1 + 20)
    #     plt.plot(noisy_input_test[i])
    #
    #     # display encoded images
    #     ax = plt.subplot(4, 20, 2 * 20 + i + 1)
    #     plt.plot(encoded_imgs[i])
    #
    #     # display reconstruction images
    #     ax = plt.subplot(4, 20, 3 * 20 + i + 1)
    #     plt.plot(predicted[i])
    #
    # plt.show()

    # Generate reconstructions
    num_reconstructions = 10
    samples = noisy_input_test[:num_reconstructions]
    reconstructions = autoencoder.predict(samples)

    # Plot reconstructions
    for i in np.arange(0, num_reconstructions):
        # Prediction index
        prediction_index = i + percentage_training
        # Get the sample and the reconstruction
        original = contaminated_parts[prediction_index]
        pure = clean_parts[prediction_index]
        reconstruction = np.array(reconstructions[i])
        # Matplotlib preparations
        fig, axes = plt.subplots(1, 3)
        # Plot sample and reconstruciton
        axes[0].plot(original)
        axes[0].set_title('Noisy waveform')
        axes[1].plot(pure)
        axes[1].set_title('Pure waveform')
        axes[2].plot(reconstruction)
        axes[2].set_title('Conv Autoencoder Denoised waveform')
        plt.show()

# predict with SVM then predict with autoencoder
